# Generated by Django 2.2.3 on 2019-09-12 20:19

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0002_auto_20190909_1707'),
    ]

    operations = [
        migrations.CreateModel(
            name='ArchivedPlaylist',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('list_id', models.CharField(max_length=200, unique=True)),
                ('title', models.CharField(max_length=1000)),
                ('counter', models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='PlaylistEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('index', models.IntegerField()),
                ('url', models.CharField(max_length=200)),
                ('playlist', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', to='core.ArchivedPlaylist')),
                ('song', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='playlist_entries', to='core.ArchivedSong')),
            ],
            options={
                'ordering': ['playlist', 'index'],
            },
        ),
        migrations.CreateModel(
            name='ArchivedPlaylistQuery',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('query', models.CharField(max_length=1000)),
                ('playlist', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='queries', to='core.ArchivedPlaylist')),
            ],
        ),
        migrations.AddField(
            model_name='requestlog',
            name='playlist',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.ArchivedPlaylist'),
        ),
    ]



# Generated by Django 2.2.3 on 2019-09-09 15:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='currentsong',
            name='manually_requested',
            field=models.BooleanField(default=True),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='playlog',
            name='manually_requested',
            field=models.BooleanField(default=True),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='queuedsong',
            name='manually_requested',
            field=models.BooleanField(default=True),
            preserve_default=False,
        ),
    ]



from django.utils import timezone
from django.conf import settings

import core.models as models

import subprocess
import ipware

class UserManager:
    last_requests = {}

    def __init__(self, base):
        self.base = base

        # kick users after some time without any request
        self.inactivity_period = 600
        self.update_user_count()

    def update_user_count(self):
        now = timezone.now()
        for key, value in list(UserManager.last_requests.items()):
            if (now - value).seconds >= self.inactivity_period:
                del UserManager.last_requests[key]
        self.last_user_count_update = now
    
    def get_count(self):
        if (timezone.now() - self.last_user_count_update).seconds >= 60:
            self.update_user_count()
        return len(UserManager.last_requests)

    def partymode_enabled(self):
        return len(UserManager.last_requests) >= self.base.settings.people_to_party
    
    def has_controls(self, user):
        return user.username == 'mod' or \
                user.username == 'pad' or \
                user.username == 'admin'

    def has_pad(self, user):
        return user.username == 'pad' or \
                user.username == 'admin'

    def is_admin(self, user):
        return user.username == 'admin'

class SimpleMiddleware:
    def __init__(self, get_response):
        # One-time configuration and initialization.
        self.get_response = get_response

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        ip, is_routable = ipware.get_client_ip(request)
        if ip is None:
            ip = ''
        UserManager.last_requests[ip] = timezone.now()

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response



from django import forms
from core.models import Pad

class TagForm(forms.Form):
    submit_tag = forms.CharField(label="tag_form_label", max_length=100, widget=forms.TextInput(attrs={'placeholder': 'Push Hashtag'}))

class PadForm(forms.Form):
    pad_field = forms.CharField(label="pad_label", widget=forms.Textarea, required=False)
    # everytime a new form is generated, it should look up the content from the database
    def __init__(self, *args, **kwargs):
        super(PadForm, self).__init__(*args, **kwargs)
        # now we can add fields using a dictionary!
        self.initial['pad_field'] = Pad.objects.all()[0].content

class MusiqForm(forms.Form):
    submit_musiq = forms.CharField(label="musiq_form_label", max_length=200)



from django.conf import settings

import core.musiq.song_utils as song_utils

import youtube_dl
import subprocess
import requests
import pickle
import errno
import json
import os

def get_initial_data(html):
    for line in html.split('\n'):
        line = line.strip()
        prefix = 'window["ytInitialData"] = '
        if line.startswith(prefix):
            # strip assignment and semicolon
            initial_data = line[len(prefix):-1]
            initial_data = json.loads(initial_data)
            return initial_data

def get_suggestion(url):
    session = requests.session()
    try:
        with open(os.path.join(settings.BASE_DIR, 'config/youtube_cookies.pickle'), 'rb') as f:
            session.cookies.update(pickle.load(f))
    except FileNotFoundError:
        pass

    headers = {
        'User-Agent': youtube_dl.utils.random_user_agent(),
    }
    r = session.get(url, headers=headers)

    with open(os.path.join(settings.BASE_DIR, 'config/youtube_cookies.pickle'), 'wb') as f:
        pickle.dump(session.cookies, f)

    initial_data = get_initial_data(r.text)
    url = initial_data['contents']['twoColumnWatchNextResults']['secondaryResults']['secondaryResults']['results'][0]['compactAutoplayRenderer']['contents'][0]['compactVideoRenderer']['navigationEndpoint']['commandMetadata']['webCommandMetadata']['url']
    return 'https://www.youtube.com' + url

def search_playlist(query):
    session = requests.session()
    try:
        with open(os.path.join(settings.BASE_DIR, 'config/youtube_cookies.pickle'), 'rb') as f:
            session.cookies.update(pickle.load(f))
    except FileNotFoundError:
        pass

    headers = {
        'User-Agent': youtube_dl.utils.random_user_agent(),
    }
    params = {
            'search_query': query,
            # this is the value that youtube uses to filter for playlists only
            'sp': 'EgQQA1AD'
    }
    r = session.get('https://www.youtube.com/results', headers=headers, params=params)

    with open(os.path.join(settings.BASE_DIR, 'config/youtube_cookies.pickle'), 'wb') as f:
        pickle.dump(session.cookies, f)

    initial_data = get_initial_data(r.text)
    search_results = initial_data['contents']['twoColumnSearchResultsRenderer']['primaryContents']['sectionListRenderer']['contents'][0]['itemSectionRenderer']['contents']

    list_id = next(res['playlistRenderer']['playlistId'] for res in search_results if 'playlistRenderer' in res)
    return list_id

class MyLogger(object):
    def debug(self, msg):
        if settings.DEBUG:
            print(msg)
    def warning(self, msg):
        if settings.DEBUG:
            print(msg)
    def error(self, msg):
        print(msg)

class SongTooLargeException(Exception):
    pass

class NoPlaylistException(Exception):
    pass

class Downloader:

    def __init__(self, musiq, target):
        self.musiq = musiq
        self.target = target
        self.info_dict = None
        # youtube-dl --format bestaudio[ext=m4a]/best[ext=m4a] --output '%(id)s.%(ext)s --no-playlist --write-thumbnail --default-search auto --add-metadata --embed-thumbnail
        self.ydl_opts = {
            'format': 'bestaudio[ext=m4a]/best[ext=m4a]',
            'outtmpl': os.path.join(settings.SONGS_CACHE_DIR, '%(id)s.%(ext)s'),
            'noplaylist': True,
            'no_color': True,
            'writethumbnail': True,
            'default_search': 'auto',
            'postprocessors': [{
                    'key': 'FFmpegMetadata',
                }, {
                    'key': 'EmbedThumbnail',
                    # overwrite any thumbnails already present
                    'already_have_thumbnail': True,
                }],
            'logger': MyLogger(),
        }

    def get_playlist_info(self):
        del self.ydl_opts['noplaylist']
        self.ydl_opts['extract_flat'] = True

        # in case of a radio playist, restrict the number of songs that are downloaded
        # if we received just the id, it is an id starting with 'RD'
        # if its a url, the id is behind a '&list='
        if song_utils.is_radio(self.target):
            self.ydl_opts['playlistend'] = self.musiq.base.settings.max_playlist_items

        with youtube_dl.YoutubeDL(self.ydl_opts) as ydl:
            self.info_dict = ydl.extract_info(self.target, download=False)

        if self.info_dict['_type'] != 'playlist' or 'entries' not in self.info_dict:
            raise NoPlaylistException('Not a Playlist')

        playlist_info = {}
        playlist_info['id'] = self.info_dict['id']
        playlist_info['urls'] = []
        if 'title' in self.info_dict:
            playlist_info['title'] = self.info_dict['title']
        for entry in self.info_dict['entries']:
            playlist_info['urls'].append('https://www.youtube.com/watch?v=' + entry['id'])
        return playlist_info

    def check(self):
        with youtube_dl.YoutubeDL(self.ydl_opts) as ydl:
            self.info_dict = ydl.extract_info(self.target, download=False)
        # this value is not an exact match, but it's a good approximation
        if 'entries' in self.info_dict:
            self.info_dict = self.info_dict['entries'][0]
        size = self.info_dict['filesize']
        max_size = self.musiq.base.settings.max_download_size * 1024 * 1024
        if max_size != 0 and song_utils.path_from_id(self.info_dict['id']) is None and (size is None or size > max_size):
            raise SongTooLargeException('Song too long')

    def fetch(self):
        # check if file was already downloaded and only download if necessary
        location = song_utils.path_from_id(self.info_dict['id'])
        if location is not None:
            return location

        with youtube_dl.YoutubeDL(self.ydl_opts) as ydl:
            ydl.download([self.target])

        location = song_utils.path_from_id(self.info_dict['id'])
        base = os.path.splitext(location)[0]
        thumbnail = base + '.jpg'
        try:
            os.remove(thumbnail)
        except FileNotFoundError:
            self.musiq.base.logger.info('tried to delete ' + thumbnail + ' but does not exist')

        try:
            # tag the file with replaygain to perform volume normalization
            subprocess.call(['aacgain', '-q', '-c', location], stdout=subprocess.DEVNULL)
        except OSError as e:
            if e.errno == errno.ENOENT:
                # the aacgain package was not found. Skip normalization
                pass
            else:
                raise
                
        return location

if __name__ == '__main__':
    Downloader().fetch()




from django.shortcuts import render
from django.core.exceptions import PermissionDenied
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import HttpResponseForbidden
from django.http import JsonResponse
from django.utils import dateparse
from django.utils import timezone
from django.db import models
from django.conf import settings

from core.models import Setting
from core.models import PlayLog
from core.models import RequestLog
import core.state_handler as state_handler
import core.musiq.song_utils as song_utils

from functools import wraps
from datetime import timedelta
from dateutil import tz
import urllib.request
import subprocess
import threading
import time
import math
import os
import re

class Settings:

    def __init__(self, base):
        self.base = base
        self.voting_system = Setting.objects.get_or_create(key='voting_system', defaults={'value': False})[0].value == 'True'
        self.logging_enabled = Setting.objects.get_or_create(key='logging_enabled', defaults={'value': True})[0].value == 'True'
        self.people_to_party = int(Setting.objects.get_or_create(key='people_to_party', defaults={'value': 3})[0].value)
        self.alarm_probability = float(Setting.objects.get_or_create(key='alarm_probability', defaults={'value': 0})[0].value)
        self.downvotes_to_kick = int(Setting.objects.get_or_create(key='downvotes_to_kick', defaults={'value': 3})[0].value)
        self.max_download_size = int(Setting.objects.get_or_create(key='max_download_size', defaults={'value': 10})[0].value)
        self.max_playlist_items = int(Setting.objects.get_or_create(key='max_playlist_items', defaults={'value': 10})[0].value)
        self._check_internet()
        self.bluetoothctl = None
        self.bluetooth_devices = []
        self.homewifi = Setting.objects.get_or_create(key='homewifi', defaults={'value': ''})[0].value

    def state_dict(self):
        state_dict = self.base.state_dict()
        state_dict['voting_system'] = self.voting_system
        state_dict['logging_enabled'] = self.logging_enabled
        state_dict['people_to_party'] = self.people_to_party
        state_dict['alarm_probability'] = self.alarm_probability
        state_dict['downvotes_to_kick'] = self.downvotes_to_kick
        state_dict['max_download_size'] = self.max_download_size
        state_dict['max_playlist_items'] = self.max_playlist_items
        state_dict['has_internet'] = self.has_internet

        state_dict['bluetooth_scanning'] = self.bluetoothctl is not None
        state_dict['bluetooth_devices'] = self.bluetooth_devices

        try:
            with open(os.path.join(settings.BASE_DIR, 'config/homewifi')) as f:
                state_dict['homewifi_ssid'] = f.read()
        except FileNotFoundError:
            state_dict['homewifi_ssid'] = ''

        try:
            state_dict['homewifi_enabled'] = subprocess.call(['/usr/local/sbin/raveberry/homewifi_enabled']) != 0
            state_dict['events_enabled'] = subprocess.call(['/usr/local/sbin/raveberry/events_enabled']) != 0
            state_dict['hotspot_enabled'] = subprocess.call(['/usr/local/sbin/raveberry/hotspot_enabled']) != 0
            state_dict['wifi_protected'] = subprocess.call(['/usr/local/sbin/raveberry/wifi_protected']) != 0
            state_dict['tunneling_enabled'] = subprocess.call(['sudo', '/usr/local/sbin/raveberry/tunneling_enabled']) != 0
            state_dict['remote_enabled'] = subprocess.call(['/usr/local/sbin/raveberry/remote_enabled']) != 0
        except FileNotFoundError:
            self.base.logger.error('scripts not installed')

        return state_dict

    def get_state(self, request):
        state = self.state_dict()
        return JsonResponse(state)

    def update_state(self):
        state_handler.update_state(self.state_dict())

    def index(self, request):
        if not self.base.user_manager.is_admin(request.user):
            raise PermissionDenied
        context = self.base.context(request)
        return render(request, 'settings.html', context)

    def _check_internet(self):
        response = subprocess.call(['ping','-c','1','-W','3','1.1.1.1'], stdout=subprocess.DEVNULL)
        if response == 0:
            self.has_internet = True
        else:
            self.has_internet = False

    # settings can only be changed by admin
    def option(func):
        def _decorator(self, request, *args, **kwargs):
            # don't allow option changes during alarm
            if request.user.username != 'admin':
                return HttpResponseForbidden()
            response = func(self, request, *args, **kwargs)
            if response is not None:
                return response
            self.update_state()
            return HttpResponse()
        return wraps(func)(_decorator)

    @option
    def set_voting_system(self, request):
        enabled = request.POST.get('value') == 'true'
        Setting.objects.filter(key='voting_system').update(value=enabled)
        self.voting_system = enabled
    @option
    def set_logging_enabled(self, request):
        enabled = request.POST.get('value') == 'true'
        Setting.objects.filter(key='logging_enabled').update(value=enabled)
        self.logging_enabled = enabled
    @option
    def set_people_to_party(self, request):
        value = int(request.POST.get('value'))
        Setting.objects.filter(key='people_to_party').update(value=value)
        self.people_to_party = value
    @option
    def set_alarm_probability(self, request):
        value = float(request.POST.get('value'))
        Setting.objects.filter(key='alarm_probability').update(value=value)
        self.alarm_probability = value
    @option
    def set_downvotes_to_kick(self, request):
        value = int(request.POST.get('value'))
        Setting.objects.filter(key='downvotes_to_kick').update(value=value)
        self.downvotes_to_kick = value
    @option
    def set_max_download_size(self, request):
        value = int(request.POST.get('value'))
        Setting.objects.filter(key='max_download_size').update(value=value)
        self.max_download_size = value
    @option
    def set_max_playlist_items(self, request):
        value = int(request.POST.get('value'))
        Setting.objects.filter(key='max_playlist_items').update(value=value)
        self.max_playlist_items = value
    @option
    def check_internet(self, request):
        self._check_internet()
    @option
    def update_user_count(self, request):
        self.base.user_manager.update_user_count()

    def _get_bluetoothctl_line(self):
        # Note: this variable is not guarded by a lock. 
        # But there should only be one admin accessing these bluetooth functions anyway.
        if self.bluetoothctl is None:
            return ''
        line = self.bluetoothctl.stdout.readline().decode()
        ansi_escape = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')
        line = ansi_escape.sub('', line)
        line = line.strip()
        return line
    def _stop_bluetoothctl(self):
        self.bluetoothctl.stdin.close()
        self.bluetoothctl.wait()
        self.bluetoothctl = None
    @option
    def set_bluetooth_scanning(self, request):
        enabled = request.POST.get('value') == 'true'
        if enabled:
            if self.bluetoothctl is not None:
                return HttpResponseBadRequest('Already Scanning')
            self.bluetooth_devices = []
            self.bluetoothctl = subprocess.Popen(["bluetoothctl"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)

            self.bluetoothctl.stdin.write(b'devices\n')
            self.bluetoothctl.stdin.write(b'scan on\n')
            self.bluetoothctl.stdin.flush()
            while True:
                line = self._get_bluetoothctl_line()
                if not line:
                    break
                # match old devices
                match = re.match('Device (\S*) (.*)', line)
                # match newly scanned devices
                # We need the '.*' at the beginning of the line to account for control sequences 
                if not match:
                    match = re.match('.*\[NEW\] Device (\S*) (.*)', line)
                if match:
                    address = match.group(1)
                    name = match.group(2)
                    # filter unnamed devices
                    # devices named after their address are no speakers
                    if re.match('[A-Z0-9][A-Z0-9](-[A-Z0-9][A-Z0-9]){5}', name):
                        continue
                    self.bluetooth_devices.append({
                        'address': address,
                        'name': name,
                    })
                    self.update_state()
        else:
            if self.bluetoothctl is None:
                return HttpResponseBadRequest('Currently not scanning')
            self._stop_bluetoothctl()
    @option
    def connect_to_bluetooth_device(self, request):
        address = request.POST.get('address')
        if self.bluetoothctl is not None:
            return HttpResponseBadRequest('Stop scanning before connecting')
        if address is None or address is '':
            return HttpResponseBadRequest('No device selected')

        self.bluetoothctl = subprocess.Popen(["bluetoothctl"], stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        error = ''

        # A Function that acts as a timeout for unexpected errors (or timeouts)
        def timeout():
            time.sleep(20)
            error = 'Timed out'
            if self.bluetoothctl is not None:
                self._stop_bluetoothctl()
        threading.Thread(target=timeout, daemon=True).start()

        self.bluetoothctl.stdin.write(b'pair ' + address.encode() + b'\n')
        self.bluetoothctl.stdin.flush()
        while True:
            line = self._get_bluetoothctl_line()
            if not line:
                break
            if re.match('.*Device ' + address + ' not available', line):
                error = 'Device unavailable'
                break
            elif re.match('.*Failed to pair: org.bluez.Error.AlreadyExists', line):
                break
            elif re.match('.*Pairing successful', line):
                break

        if error:
            self._stop_bluetoothctl()
            return HttpResponseBadRequest(error)

        self.bluetoothctl.stdin.write(b'connect ' + address.encode() + b'\n')
        self.bluetoothctl.stdin.flush()
        while True:
            line = self._get_bluetoothctl_line()
            if not line:
                break
            if re.match('.*Device ' + address + ' not available', line):
                error = 'Device unavailable'
                break
            elif re.match('.*Failed to connect: org.bluez.Error.Failed', line):
                error = 'Connect Failed'
                break
            elif re.match('.*Failed to connect: org.bluez.Error.InProgress', line):
                error = 'Connect in progress'
                break
            elif re.match('.*Connection successful', line):
                break
        # trust the device to automatically reconnect when it is available again
        self.bluetoothctl.stdin.write(b'trust ' + address.encode() + b'\n')
        self.bluetoothctl.stdin.flush()

        self._stop_bluetoothctl()
        if error:
            return HttpResponseBadRequest(error)

        # Update mpd's config to output to the bluetooth device
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/update_bluetooth_device', address])
        return HttpResponse('Connected')

    @option
    def available_ssids(self, request):
        output = subprocess.check_output(['sudo', '/usr/local/sbin/raveberry/list_available_ssids'])
        output = output.decode()
        ssids = output.split('\n')
        print(ssids)
        return JsonResponse(ssids[:-1], safe=False)
    @option
    def connect_to_wifi(self, request):
        ssid = request.POST.get('ssid')
        password = request.POST.get('password')
        if ssid is None or password is None \
                or ssid == '' or password == '':
            return HttpResponseBadRequest('Please provide both SSID and password')
        try:
            output = subprocess.check_output(['sudo', '/usr/local/sbin/raveberry/connect_to_wifi', ssid, password])
            output = output.decode()
            return HttpResponse(output)
        except subprocess.CalledProcessError as e:
            output = e.output.decode()
            return HttpResponseBadRequest(output)
    @option
    def disable_homewifi(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/disable_homewifi'])
    @option
    def enable_homewifi(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/enable_homewifi'])
    @option
    def stored_ssids(self, request):
        output = subprocess.check_output(['sudo', '/usr/local/sbin/raveberry/list_stored_ssids'])
        output = output.decode()
        ssids = output.split('\n')
        return JsonResponse(ssids[:-1], safe=False)
    @option
    def set_homewifi_ssid(self, request):
        homewifi_ssid = request.POST.get('homewifi_ssid')
        with open(os.path.join(settings.BASE_DIR, 'config/homewifi'), 'w+') as f:
            f.write(homewifi_ssid)

    @option
    def analyse(self, request):
        startdate = request.POST.get('startdate')
        starttime = request.POST.get('starttime')
        enddate = request.POST.get('enddate')
        endtime = request.POST.get('endtime')
        if startdate is None or startdate == '' \
                or starttime is None or starttime == '' \
                or enddate is None or enddate == '' \
                or endtime is None or endtime == '':
            return HttpResponseBadRequest('All fields are required')

        start = dateparse.parse_datetime(startdate + 'T' + starttime)
        end = dateparse.parse_datetime(enddate + 'T' + endtime)

        if start is None or end is None:
            return HttpResponseBadRequest('invalid start-/endtime given')
        if start >= end:
            return HttpResponseBadRequest('start has to be before end')

        start = timezone.make_aware(start)
        end = timezone.make_aware(end)

        played = PlayLog.objects.all().filter(created__gte=start).filter(created__lt=end)
        requested = RequestLog.objects.all().filter(created__gte=start).filter(created__lt=end)
        played_count = played.values('song__url', 'song__artist', 'song__title').values('song__url', 'song__artist', 'song__title', count=models.Count('song__url')).order_by('-count')
        played_votes = PlayLog.objects.all().filter(created__gte=start).filter(created__lt=end).order_by('-votes')
        devices = requested.values('address').values('address', count=models.Count('address'))
        
        response = {}
        response['songs_played'] = len(played);
        response['most_played_song'] = song_utils.displayname(
                played_count[0]['song__artist'],
                played_count[0]['song__title']) + ' (' + str(played_count[0]['count']) + ')'
        response['highest_voted_song'] = played_votes[0].song.displayname() + ' (' + str(played_votes[0].votes) + ')'
        response['most_active_device'] = devices[0]['address'] + ' (' + str(devices[0]['count']) + ')'
        requested_by_ip = requested.filter(address=devices[0]['address'])
        for i in range(6):
            if i >= len(requested_by_ip):
                break
            response['most_active_device'] += '\n'
            if i == 5:
                response['most_active_device'] += '...'
            else:
                response['most_active_device'] += requested_by_ip[i].song.displayname()


        binsize = 3600
        number_of_bins = math.ceil((end - start).total_seconds() / binsize)
        request_bins = [0 for _ in range(number_of_bins)]

        for r in requested:
            seconds = (r.created - start).total_seconds()
            index = int(seconds / binsize)
            request_bins[index] += 1

        current_time = start
        current_index = 0
        response['request_activity'] = ''
        while current_time < end:
            response['request_activity'] += current_time.strftime('%H:%M') 
            response['request_activity'] += ':\t' + str(request_bins[current_index])
            response['request_activity'] += '\n'
            current_time += timedelta(seconds=binsize)
            current_index += 1

        localtz = tz.gettz(settings.TIME_ZONE)
        playlist = ''
        for log in played:
            localtime = log.created.astimezone(localtz)
            playlist += '[{:02d}:{:02d}] {}\n'.format(localtime.hour, localtime.minute, log.song.displayname())
        response['playlist'] = playlist

        return JsonResponse(response)

    @option
    def disable_events(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/disable_events'])
    @option
    def enable_events(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/enable_events'])
    @option
    def disable_hotspot(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/disable_hotspot'])
    @option
    def enable_hotspot(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/enable_hotspot'])
    @option
    def unprotect_wifi(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/unprotect_wifi'])
    @option
    def protect_wifi(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/protect_wifi'])
    @option
    def disable_tunneling(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/disable_tunneling'])
    @option
    def enable_tunneling(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/enable_tunneling'])
    @option
    def disable_remote(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/disable_remote'])
    @option
    def enable_remote(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/enable_remote'])
    @option
    def reboot_server(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/reboot_server'])
    @option
    def reboot_system(self, request):
        subprocess.call(['sudo', '/usr/local/sbin/raveberry/reboot_system'])



from django.db import transaction
from django.db.models import Q
from django.db.models import F
from django.http import HttpResponse
from django.http import HttpResponseBadRequest
from django.http import JsonResponse

from core.musiq.youtube import Downloader
from core.musiq.youtube import NoPlaylistException
import core.musiq.youtube as youtube
from core.models import ArchivedSong
from core.models import ArchivedPlaylist
from core.models import ArchivedQuery
from core.models import ArchivedPlaylistQuery
from core.models import PlaylistEntry
from core.models import RequestLog
import core.musiq.song_utils as song_utils

import json
import random

class OfflineDownloader():
    def __init__(self, location):
        self.location = location
    def fetch(self):
        return self.location

class SongProvider:

    def __init__(self, musiq):
        self.musiq = musiq

    def check_archived_song_accessible(self, key):
        archived_song = ArchivedSong.objects.get(id=key)
        location = song_utils.path_from_url(archived_song.url)
        downloader = None
        if not location:
            downloader = Downloader(self.musiq, archived_song.url)
            downloader.check()
        # the downloader raised no error, the song is accessible
        return downloader

    def get_archived_song_location(self, key, downloader, ip, archive=True):
        archived_song = ArchivedSong.objects.get(id=key)
        location = song_utils.path_from_url(archived_song.url)
        if location is None:
            location = downloader.fetch()
        if archive:
            ArchivedSong.objects.filter(id=key).update(counter=F('counter')+1)
        if archive and self.musiq.base.settings.logging_enabled:
            RequestLog.objects.create(song=archived_song, address=ip)
        return location

    def check_new_song_accessible(self, search_text):
        # collapse whitespaces
        search_text = ' '.join(search_text.split())

        # if the search text is a url that is already in the database and the song is cached offline, we don't need an online downloader
        downloader = None
        queryset = ArchivedSong.objects.filter(url=search_text)
        if queryset.count() == 1:
            song = queryset.get()
            location = song_utils.path_from_url(song.url)
            if location is not None:
                downloader = OfflineDownloader(location)

        # if the song was not found offline, we use the online downloader
        if downloader is None:
            downloader = Downloader(self.musiq, search_text)
            downloader.check()
            # the downloader raised no error, the song is accessible
        return downloader

    def get_new_song_location(self, search_text, downloader, ip, archive=True):
        location = downloader.fetch()

        metadata = song_utils.gather_metadata(location)

        with transaction.atomic():
            queryset = ArchivedSong.objects.filter(url=metadata['url'])
            if queryset.count() == 0:
                initial_counter = 1 if archive else 0
                archived_song = ArchivedSong.objects.create(url=metadata['url'], artist=metadata['artist'], title=metadata['title'], counter=initial_counter)
            else:
                if archive:
                    queryset.update(counter=F('counter')+1)
                archived_song = queryset.get()

            if archive:
                ArchivedQuery.objects.get_or_create(song=archived_song, query=search_text)

        if archive and self.musiq.base.settings.logging_enabled:
            RequestLog.objects.create(song=archived_song, address=ip)
        return location

    def get_new_playlist(self, query, ip):
        try:
            downloader = Downloader(self.musiq, query)
            playlist_info = downloader.get_playlist_info()
        except NoPlaylistException:
            # no valid playlist url was provided. use the query as a search
            list_id = youtube.search_playlist(query)
            downloader = Downloader(self.musiq, list_id)
            playlist_info = downloader.get_playlist_info()

        if playlist_info['id'].startswith('RD'):
            # the radio for a given song can be different every time it is accessed
            # thus, we create a new database entry every time a radio is requested
            radio = True
        else:
            radio = False

        with transaction.atomic():
            queryset = ArchivedPlaylist.objects.filter(list_id=playlist_info['id'])
            if radio or queryset.count() == 0:
                archived_playlist = ArchivedPlaylist.objects.create(list_id=playlist_info['id'], title=playlist_info['title'], counter=1)
                for index, url in enumerate(playlist_info['urls']):
                    PlaylistEntry.objects.create(
                        playlist=archived_playlist,
                        index=index,
                        url=url,
                    )
            else:
                queryset.update(counter=F('counter')+1)
                archived_playlist = queryset.get()

            ArchivedPlaylistQuery.objects.get_or_create(playlist=archived_playlist, query=query)

        if self.musiq.base.settings.logging_enabled:
            RequestLog.objects.create(playlist=archived_playlist, address=ip)

        return archived_playlist

    def get_archived_playlist(self, key, ip):
        archived_playlist = ArchivedPlaylist.objects.get(id=key)
        ArchivedPlaylist.objects.filter(id=key).update(counter=F('counter')+1)
        if self.musiq.base.settings.logging_enabled:
            RequestLog.objects.create(playlist=archived_playlist, address=ip)
        return archived_playlist

    def random_suggestion(self, request):
        suggest_playlist = request.GET['playlist'] == 'true'
        if suggest_playlist:
            # exclude radios from suggestions
            remaining_playlists = ArchivedPlaylist.objects.all().exclude(list_id__startswith='RD').exclude(list_id__contains='&list=RD')
            if remaining_playlists.count() == 0:
                return HttpResponseBadRequest('No playlists to suggest from')
            index = random.randint(0, remaining_playlists.count() - 1)
            playlist = remaining_playlists.all()[index]
            return JsonResponse({
                'suggestion': playlist.title,
                'key': playlist.id,
            })
        else:
            if ArchivedSong.objects.count() == 0:
                return HttpResponseBadRequest('No songs to suggest from')
            index = random.randint(0,ArchivedSong.objects.count() - 1)
            song = ArchivedSong.objects.all()[index]
            return JsonResponse({
                'suggestion': song.displayname(),
                'key': song.id,
            })

    def get_suggestions(self, request):
        terms = request.GET['term'].split()
        suggest_playlist = request.GET['playlist'] == 'true'

        results = []
        if suggest_playlist:
            remaining_playlists = ArchivedPlaylist.objects.prefetch_related('queries')
            # exclude radios from suggestions
            remaining_playlists = remaining_playlists.exclude(list_id__startswith='RD').exclude(list_id__contains='&list=RD')

            for term in terms:
                remaining_playlists = remaining_playlists.filter(Q(title__icontains=term) | Q(queries__query__icontains=term))

            remaining_playlists = remaining_playlists \
                .values('id', 'title', 'counter') \
                .distinct() \
                .order_by('-counter') \
                [:20]

            for playlist in remaining_playlists:
                cached = False
                result_dict = {
                    'key': playlist['id'],
                    'value': playlist['title'],
                    'counter': playlist['counter'],
                    'type': 'cached' if cached else 'online',
                }
                results.append(result_dict)
        else:
            remaining_songs = ArchivedSong.objects.prefetch_related('queries')

            for term in terms:
                remaining_songs = remaining_songs.filter(Q(title__icontains=term) | Q(artist__icontains=term) | Q(queries__query__icontains=term))

            remaining_songs = remaining_songs \
                .values('id', 'title', 'url', 'artist', 'counter') \
                .distinct() \
                .order_by('-counter') \
                [:20]

            for song in remaining_songs:
                if song_utils.path_from_url(song['url']) is not None:
                    cached = True
                else:
                    cached = False
                # don't suggest online songs when we don't have internet
                if not self.musiq.base.settings.has_internet:
                    if not cached:
                        continue
                result_dict = {
                    'key': song['id'],
                    'value': song_utils.displayname(song['artist'], song['title']),
                    'counter': song['counter'],
                    'type': 'cached' if cached else 'online',
                }
                results.append(result_dict)

        return HttpResponse(json.dumps(results))

""" query for the suggestions
SELECT DISTINCT id, title, artist, counter
FROM core_archivedsong s LEFT JOIN core_archivedquery q ON q.song
WHERE forall term in terms: term in q.query or term in s.artist or term in s.title
ORDER BY -counter
"""



from django.contrib import admin

import core.models as models

# Register your models here.
admin.site.register(models.Tag)
admin.site.register(models.Counter)
admin.site.register(models.Pad)
admin.site.register(models.QueuedSong)
admin.site.register(models.CurrentSong)
admin.site.register(models.ArchivedSong)
admin.site.register(models.ArchivedPlaylist)
admin.site.register(models.PlaylistEntry)
admin.site.register(models.ArchivedQuery)
admin.site.register(models.ArchivedPlaylistQuery)
admin.site.register(models.Setting)
admin.site.register(models.RequestLog)
admin.site.register(models.PlayLog)



from django.conf import settings

from urllib.parse import urlparse
from urllib.parse import parse_qs
import os
import mutagen.easymp4

def path_from_id(song_id):
    path = os.path.join(settings.SONGS_CACHE_DIR, song_id + '.m4a')
    path = path.replace('~', os.environ['HOME'])
    path = os.path.abspath(path)
    if os.path.isfile(path):
        return path
    else:
        return None

def path_from_url(url):
    ''' returns the file path for the song corresponding to the given url or None if the url is not cached '''
    song_id = id_from_url(url)
    return path_from_id(song_id)

def id_from_url(url):
    return parse_qs(urlparse(url).query)['v'][0]

def is_radio(id_or_url):
    return id_or_url.split('&list=')[-1].startswith('RD')

def format_seconds(seconds):
    hours, seconds =  seconds // 3600, seconds % 3600
    minutes, seconds = seconds // 60, seconds % 60

    formatted = ''
    if hours > 0:
        formatted += '{:02d}:'.format(int(hours))
    formatted += '{0:02d}:{1:02d}'.format(int(minutes), int(seconds))
    return formatted

def get_duration(location):
    parsed = mutagen.File(location)
    return parsed.info.length

def gather_metadata(location):
    '''gathers the metadata for the song at the given location.
    'title' and 'duration' is read from tags, the 'url' is built from the location'''

    parsed = mutagen.easymp4.EasyMP4(location)
    metadata = {}
    metadata['url'] = 'https://www.youtube.com/watch?v=' + os.path.splitext(os.path.basename(location))[0]
    if parsed.tags is not None:
        if 'artist' in parsed.tags:
            metadata['artist'] = parsed.tags['artist'][0]
        if 'title' in parsed.tags:
            metadata['title'] = parsed.tags['title'][0]
    if 'artist' not in metadata:
        metadata['artist'] = ''
    if 'title' not in metadata:
        metadata['title'] = metadata['url']
    if parsed.info is not None and parsed.info.length is not None:
        metadata['duration'] = format_seconds(parsed.info.length)
    else:
        metadata['duration'] = '??:??'
    
    return metadata

def decide_deletion(location):
    return
    ''' takes the path to a song file and deletes it if it is larger than a threshold '''
    '''size = os.path.getsize(location)
    if size > settings.MAX_SONG_SIZE:
        os.remove(location)'''

def displayname(artist, title):
    if artist == '':
        return title
    else:
        return artist + ' – ' + title



from django.conf import settings
from django.db import models
from django.db import transaction
from django.db.models import F

import core.musiq.song_utils as song_utils
import core.models

class SongQueue(models.Manager):

    @transaction.atomic
    def enqueue(self, location, manually_requested):
        metadata = song_utils.gather_metadata(location)

        last = self.last()
        index = 1 if last is None else last.index + 1
        song = self.create(
                index=index,
                manually_requested=manually_requested,
                url=metadata['url'],
                artist=metadata['artist'],
                title=metadata['title'],
                duration=metadata['duration'])
        return song

    @transaction.atomic
    def dequeue(self):
        first = self.first()
        first_id = first.id
        if first is not None:
            first.delete()
        self.update(index=F('index')-1)
        return first_id, first

    @transaction.atomic
    def prioritize(self, key):
        to_prioritize = self.get(id=key)
        first = self.first()
        if to_prioritize == first:
            return

        self.filter(index__lt=to_prioritize.index).update(index=F('index')+1)
        to_prioritize.index = 1
        to_prioritize.save()


    @transaction.atomic
    def remove(self, key):
        to_remove = self.get(id=key)
        to_remove.delete()
        if next is not None:
            self.filter(index__gt=to_remove.index).update(index=F('index')-1)
        return to_remove

    @transaction.atomic
    def reorder(self, new_prev_id, element_id, new_next_id):

        try:
            new_prev = self.get(id=new_prev_id)
        except core.models.QueuedSong.DoesNotExist:
            new_prev = None
        try:
            to_reorder = self.get(id=element_id)
        except core.models.QueuedSong.DoesNotExist:
            raise ValueError('reordered song does not exist')
        try:
            new_next = self.get(id=new_next_id)
        except core.models.QueuedSong.DoesNotExist:
            new_next = None

        first = self.first()
        last = self.last()
        # check validity of request
        if new_prev is None and new_next is None:
            # to_reorder has to be the only element in the queue
            if first != to_reorder or last != to_reorder:
                raise ValueError('reordered song is not the only one')
        if new_prev is None and new_next is not None:
            # new_next has to be the first element
            if new_next != first:
                raise ValueError('given first is not head of the queue')
        if new_prev is not None and new_next is None:
            # new_prev has to be the last element
            if new_prev != last:
                raise ValueError('given last is not tail of the queue')
        if new_prev is not None and new_next is not None:
            # new_prev and new_next have to be adjacent
            if new_next.index != new_prev.index + 1:
                raise ValueError('given pair of songs is not adjacent')

        try:
            old_prev = self.get(index=to_reorder.index-1)
        except core.models.QueuedSong.DoesNotExist:
            old_prev = None
        try:
            old_next = self.get(index=to_reorder.index+1)
        except core.models.QueuedSong.DoesNotExist:
            old_next = None

        # update indices
        up = True
        if new_prev is not None:
            if new_prev.index > to_reorder.index:
                up = False
        elif new_next is not None:
            if new_next.index > to_reorder.index:
                up = False

        to_update = self.all()
        if up:
            if old_prev is not None:
                to_update = to_update.filter(index__lte=old_prev.index)
            new_index = 1
            if new_next is not None:
                new_index = new_next.index
                to_update = to_update.filter(index__gte=new_next.index)
            to_update.update(index=F('index')+1)
        else:
            new_index = 1
            if new_prev is not None:
                new_index = new_prev.index
                to_update = to_update.filter(index__lte=new_prev.index)
            if old_next is not None:
                to_update = to_update.filter(index__gte=old_next.index)
            to_update.update(index=F('index')-1)

        to_reorder.index = new_index
        to_reorder.save()

    @transaction.atomic
    def vote_up(self, key):
        self.filter(id=key).update(votes=F('votes')+1)

    @transaction.atomic
    def vote_down(self, key, threshold):
        self.filter(id=key).update(votes=F('votes')-1)
        try:
            song = self.get(id=key)
            if song.votes <= threshold:
                song.delete()
                return song
        except core.models.QueuedSong.DoesNotExist:
            pass
        return None



from django.conf import settings
from django.shortcuts import render
from django.http import HttpResponse
from django.http import JsonResponse
from django.http import HttpResponseBadRequest
from django.http import HttpResponseServerError
from django.core import serializers
from django.forms.models import model_to_dict

from core.models import QueuedSong
from core.models import CurrentSong
from core.models import ArchivedSong
from core.musiq.song_provider import SongProvider
from core.musiq.player import Player
from core.musiq.song_queue import SongQueue
from core.musiq.youtube import SongTooLargeException
from core.musiq.youtube import NoPlaylistException
import core.musiq.song_utils as song_utils
import core.state_handler as state_handler

import youtube_dl
import re
import threading

import time
import logging
import ipware

class Musiq:

    def __init__(self, base):
        self.base = base

        self.logger = logging.getLogger('raveberry')

        self.song_provider = SongProvider(self)

        self.queue = QueuedSong.objects
        self.placeholders = []

        self.player = Player(self)
        self.player.start()

    def request_song(self, request, query, check_function, get_function, key_or_query, archive=True, background_download=True):
        # only get ip on user requests
        if request is not None:
            ip, is_routable = ipware.get_client_ip(request)
            if ip is None:
                ip = ''
        else:
            ip = ''

        try:
            downloader = check_function(key_or_query)
        except (youtube_dl.utils.DownloadError, SongTooLargeException) as e:
            self.logger.info('video not accessible: ' + str(key_or_query))
            self.logger.info(e)
            return HttpResponseBadRequest(e)

        placeholder = {'query': query, 'replaced_by': None}
        self.placeholders.append(placeholder)
        self.update_state()

        def get_song():
            error = None
            location = None
            try:
                location = get_function(key_or_query, downloader, ip, archive=archive)
            except youtube_dl.utils.DownloadError as e:
                error = e
            
            if error is not None or location is None:
                self.logger.error('accessible video could not be downloaded: ' + str(key_or_query))
                self.logger.error(error)
                self.logger.error('location: ' + str(location))
                self.placeholders.remove(placeholder);
                self.update_state()
                return

            # if there is an actual request object, it was initiated by a user
            manually_requested = request is not None
            song = self.queue.enqueue(location, manually_requested)
            placeholder['replaced_by'] = song.id
            self.update_state()
            Player.queue_semaphore.release()
        
        thread = threading.Thread(target=get_song, daemon=True)
        thread.start()
        if not background_download:
            thread.join()

        return HttpResponse('Song queued')

    def request_playlist(self, request, get_function, key_or_query):
        ip, is_routable = ipware.get_client_ip(request)
        if ip is None:
            ip = ''

        try:
            playlist = get_function(key_or_query, ip)
        except (youtube_dl.utils.DownloadError, NoPlaylistException) as e:
            self.logger.info('playlist not accessible: ' + str(key_or_query))
            self.logger.info(e)
            return HttpResponseBadRequest(e)

        def get_playlist():
            for index, entry in enumerate(playlist.entries.all()):
                if index == self.base.settings.max_playlist_items:
                    break
                # request every url in the playlist as their own url
                response = self.request_song(None, entry.url, self.song_provider.check_new_song_accessible, self.song_provider.get_new_song_location, entry.url, archive=False, background_download=False)
                if settings.DEBUG:
                    # the sqlite database has problems if songs are pushed very fast while a new song is taken from the queue. Add a delay to mitigate.
                    time.sleep(1)
                # after the download finished successfully, the song definitely exists in the database. Now we can update the reference in the playlist entry.
                # after an error the song was not pushed and there exists no database entry. Then, we skip the update and leave the field empty.
                if type(response) == HttpResponse:
                    queryset = ArchivedSong.objects.filter(url=entry.url)
                    song = queryset.get()
                    entry.song = song
                    entry.save()
        threading.Thread(target=get_playlist, daemon=True).start()

        return HttpResponse('Queuing playlist')

    def request_archived_music(self, request):
        key = request.POST.get('key')
        playlist = request.POST.get('playlist') == 'true'
        if key is None:
            return HttpResponseBadRequest()
        query = request.POST.get('query')

        if playlist:
            return self.request_playlist(request, self.song_provider.get_archived_playlist, key)
        else:
            return self.request_song(request, query, self.song_provider.check_archived_song_accessible, self.song_provider.get_archived_song_location, key)

    def request_new_music(self, request):
        query = request.POST.get('query')
        playlist = request.POST.get('playlist') == 'true'
        if query is None or query == '':
            return HttpResponseBadRequest()

        if playlist:
            return self.request_playlist(request, self.song_provider.get_new_playlist, query)
        else:
            return self.request_song(request, query, self.song_provider.check_new_song_accessible, self.song_provider.get_new_song_location, query)

    def request_radio(self, request):
        try:
            current_song = CurrentSong.objects.get()
        except CurrentSong.DoesNotExist:
            return HttpResponseBadRequest('Need a song to play the radio')
        song_id = song_utils.id_from_url(current_song.url)
        radio_id = 'RD' + song_id
        response = self.request_playlist(request, self.song_provider.get_new_playlist, radio_id)
        if type(response) == HttpResponse:
            return HttpResponse('Queuing radio')
        else:
            return response

    def index(self, request):
        context = self.base.context(request)
        return render(request, 'musiq.html', context)

    def state_dict(self):
        state_dict = self.base.state_dict()
        try:
            current_song = CurrentSong.objects.get()
            current_song = model_to_dict(current_song)
        except CurrentSong.DoesNotExist:
            current_song = None
        song_queue = []
        all_songs = self.queue.all()
        if self.base.settings.voting_system:
            all_songs = all_songs.order_by('-votes', 'index')
        for song in all_songs:
            song_dict = model_to_dict(song)
            song_dict['confirmed'] = True
            # find the query of the placeholder that this song replaces (if any)
            for i, placeholder in enumerate(self.placeholders[:]):
                if placeholder['replaced_by'] == song.id:
                    song_dict['replaces'] = placeholder['query']
                    self.placeholders.remove(placeholder)
                    break
            else:
                song_dict['replaces'] = None
            song_queue.append(song_dict)
        song_queue += [{'title': placeholder['query'], 'confirmed': False} for placeholder in self.placeholders]

        state_dict['current_song'] =  current_song
        state_dict['paused'] =  self.player.paused()
        state_dict['progress'] =  self.player.progress()
        state_dict['shuffle'] =  self.player.shuffle
        state_dict['repeat'] =  self.player.repeat
        state_dict['autoplay'] =  self.player.autoplay
        state_dict['volume'] =  self.player.volume
        state_dict['song_queue'] =  song_queue
        return state_dict

    def get_state(self, request):
        state = self.state_dict()
        return JsonResponse(state)

    def update_state(self):
        state_handler.update_state(self.state_dict())



from django.db import transaction
from django.db.models import F
from django.http import HttpResponse
from django.http import HttpResponseForbidden
from django.http import HttpResponseBadRequest
from django.utils import timezone
from django.conf import settings

from core.models import Setting
import core.models as models
import core.musiq.song_utils as song_utils
import core.musiq.youtube as youtube

from threading import Semaphore
from threading import Lock
from threading import Event
from threading import Thread
from datetime import datetime
from functools import wraps
from contextlib import contextmanager
import os
import mpd
import time
import random

class Player:
    queue_semaphore = None

    def __init__(self, musiq):
        self.SEEK_DISTANCE = 10
        self.shuffle = Setting.objects.get_or_create(key='shuffle', defaults={'value': False})[0].value == 'True'
        self.repeat = Setting.objects.get_or_create(key='repeat', defaults={'value': False})[0].value == 'True'
        self.autoplay = Setting.objects.get_or_create(key='autoplay', defaults={'value': False})[0].value == 'True'

        self.musiq = musiq
        self.queue = models.QueuedSong.objects
        Player.queue_semaphore = Semaphore(self.queue.count())
        self.alarm_playing = Event()

        self.player = mpd.MPDClient()
        self.player_lock = Lock()
        with self.mpd_command(important=True):
            self.player.clear()

        with self.mpd_command(important=True):
            status = self.player.status()
            currentsong = self.player.currentsong()
            self.volume = int(status['volume']) / 100

    def start(self):
        Thread(target=self._loop, daemon=True).start()

    def progress(self):
        # the state is either pause or stop
        status = {}
        currentsong = {}
        with self.mpd_command() as allowed:
            if allowed:
                status = self.player.status()
                currentsong = self.player.currentsong()
        if 'elapsed' not in status or 'time' not in currentsong:
            return 0
        return 100 * float(status['elapsed']) / float(currentsong['time'])
    def paused(self):
        # the state is either pause or stop
        paused = False
        with self.mpd_command() as allowed:
            if allowed:
                paused = self.player.status()['state'] != 'play'
        return paused

    def _loop(self):
        while True:

            catch_up = None
            if models.CurrentSong.objects.exists():
                # recover interrupted song from database
                current_song = models.CurrentSong.objects.get()

                # continue with the current song (approximately) where we last left
                duration = song_utils.get_duration(current_song.location)
                catch_up = (timezone.now() - current_song.created).total_seconds()
            else:
                self.queue_semaphore.acquire()

                # select the next song depending on settings
                if self.musiq.base.settings.voting_system:
                    with transaction.atomic():
                        song = self.queue.all().order_by('-votes', 'index')[0]
                        song_id = song.id
                        self.queue.remove(song.id)
                elif self.shuffle:
                    index = random.randint(0,models.QueuedSong.objects.count() - 1)
                    song_id = models.QueuedSong.objects.all()[index].id
                    song = self.queue.remove(song_id)
                else:
                    # move the first song in the queue into the current song
                    song_id, song = self.queue.dequeue()

                if song is None:
                    # either the semaphore didn't match up with the actual count of songs in the queue or a race condition occured
                    self.musiq.base.logger.info('dequeued on empty list')
                    continue
                
                location = song_utils.path_from_url(song.url)
                current_song = models.CurrentSong.objects.create(
                        queue_key=song_id,
                        manually_requested=song.manually_requested,
                        votes=song.votes,
                        url=song.url,
                        artist=song.artist,
                        title=song.title,
                        duration=song.duration,
                        location=location)

                self._handle_autoplay()

                try:
                    archived_song = models.ArchivedSong.objects.get(url=current_song.url)
                    if self.musiq.base.settings.voting_system:
                        votes = current_song.votes
                    else:
                        votes = None
                    if self.musiq.base.settings.logging_enabled:
                        models.PlayLog.objects.create(
                                song=archived_song,
                                manually_requested=current_song.manually_requested,
                                votes=votes)
                except (models.ArchivedSong.DoesNotExist, models.ArchivedSong.MultipleObjectsReturned):
                    pass

            self.musiq.update_state()
        
            with self.mpd_command(important=True):
                self.player.add('file://' + current_song.location)
                self.player.play()
                if catch_up is not None:
                    self.player.seekcur(catch_up)
            self.musiq.update_state()

            self._wait_until_song_end()

            with self.mpd_command(important=True):
                try:
                    self.player.delete(0)
                except mpd.base.CommandError:
                    # catch Bad song index if there is no song
                    pass
            current_song.delete()

            if self.repeat:
                self.queue.enqueue(current_song.location, current_song.manually_requested)
                self.queue_semaphore.release()
            else:
                # the song left the queue, we can delete big downloads
                song_utils.decide_deletion(current_song.location)

            self.musiq.update_state()

            if self.musiq.base.user_manager.partymode_enabled() and random.random() < self.musiq.base.settings.alarm_probability:
                self.alarm_playing.set()
                self.musiq.base.lights.alarm_started()

                with self.mpd_command(important=True):
                    self.player.add('file://'+os.path.join(settings.BASE_DIR, 'config/sounds/alarm.m4a'))
                    self.player.play()
                self._wait_until_song_end()

                self.musiq.base.lights.alarm_stopped()
                with self.mpd_command(important=True):
                    self.player.delete(0)
                self.alarm_playing.clear()

    def _wait_until_song_end(self):
        # wait until the song is over
        while True:
            with self.mpd_command() as allowed:
                if allowed:
                    try:
                        if self.player.status()['state'] == 'stop':
                            break
                    except mpd.base.ConnectionError:
                        pass
            time.sleep(0.1) 

    def _handle_autoplay(self, url=None):
        if self.autoplay and models.QueuedSong.objects.count() == 0:
            if url is None:
                # if no url was specified, use the one of the current song
                try:
                    current_song = models.CurrentSong.objects.get()
                    url = current_song.url
                except (models.CurrentSong.DoesNotExist, models.CurrentSong.MultipleObjectsReturned):
                    return

            try:
                suggestion = youtube.get_suggestion(url)
            except Exception as e:
                self.musiq.base.logger.error('error during suggestions for ' + url)
                self.musiq.base.logger.error(e)
            else:
                self.musiq.request_song(
                        None,
                        suggestion, 
                        self.musiq.song_provider.check_new_song_accessible,
                        self.musiq.song_provider.get_new_song_location,
                        suggestion,
                        archive=False)



    # wrapper method for our mpd client that pings the mpd server before any command and reconnects if necessary. also catches protocol errors
    @contextmanager
    def mpd_command(self, important=False):
        timeout = 3
        if important:
            timeout = -1
        if self.player_lock.acquire(timeout=timeout):
            try:
                self.player.ping()
            except (mpd.base.ConnectionError, ConnectionResetError):
                for _ in range(5):
                    try:
                        self.player.connect('/var/run/mpd/socket', 6600)
                        break
                    except FileNotFoundError:
                        # system mpd is not running, try user mpd
                        try:
                            self.player.connect(os.path.expanduser('~/.mpd/socket'), 6600)
                        except mpd.base.ConnectionError:
                            # Already connected
                            break
                    except mpd.base.ConnectionError:
                        # Already connected
                        break
                    except (ConnectionResetError, ConnectionRefusedError):
                        time.sleep(0.5)
            except mpd.base.ProtocolError as e:
                print('protocol error during ping, continuing')
                print(e)
                pass
            try:
                yield True
            except mpd.base.ProtocolError as e:
                print(e)
                raise e
            finally:
                self.player_lock.release()
        else:
            yield False

    # every control changes the views state and returns an empty response
    def control(func):
        def _decorator(self, request, *args, **kwargs):
            # don't allow controls during alarm
            if self.alarm_playing.is_set():
                return HttpResponseBadRequest()
            func(self, request, *args, **kwargs)
            self.musiq.update_state()
            return HttpResponse()
        return wraps(func)(_decorator)
    # in the voting system only the admin can control the player
    def disabled_when_voting(func):
        def _decorator(self, request, *args, **kwargs):
            if self.musiq.base.settings.voting_system and not self.musiq.base.user_manager.has_controls(request.user):
                return HttpResponseForbidden()
            func(self, request, *args, **kwargs)
            self.musiq.update_state()
            return HttpResponse()
        return wraps(func)(_decorator)

    @disabled_when_voting
    @control
    def restart(self, request):
        with self.mpd_command() as allowed:
            if allowed:
                if self.player.status()['state'] != 'stop':
                    self.player.seekcur(0)
    @disabled_when_voting
    @control
    def seek_backward(self, request):
        with self.mpd_command() as allowed:
            if allowed:
                status = self.player.status()
                if 'elapsed' in status:
                    self.player.seekcur(float(status['elapsed']) - self.SEEK_DISTANCE)
    @disabled_when_voting
    @control
    def play(self, request):
        # unlikely race condition possible. but its impact is not worth the effort to prevent it
        with self.mpd_command() as allowed:
            if allowed:
                if self.player.status()['state'] == 'pause':
                    self.player.pause()
    @disabled_when_voting
    @control
    def pause(self, request):
        with self.mpd_command() as allowed:
            if allowed:
                if self.player.status()['state'] == 'play':
                    self.player.pause()
    @disabled_when_voting
    @control
    def seek_forward(self, request):
        with self.mpd_command() as allowed:
            if allowed:
                status = self.player.status()
                if 'elapsed' in status:
                    self.player.seekcur(float(status['elapsed']) + self.SEEK_DISTANCE)
    @disabled_when_voting
    @control
    def skip(self, request):
        with self.mpd_command() as allowed:
            if allowed:
                self.player.stop()
    @disabled_when_voting
    @control
    def set_shuffle(self, request):
        enabled = request.POST.get('value') == 'true'
        Setting.objects.filter(key='shuffle').update(value=enabled)
        self.shuffle = enabled
    @disabled_when_voting
    @control
    def set_repeat(self, request):
        enabled = request.POST.get('value') == 'true'
        Setting.objects.filter(key='repeat').update(value=enabled)
        self.repeat = enabled
    @disabled_when_voting
    @control
    def set_autoplay(self, request):
        enabled = request.POST.get('value') == 'true'
        Setting.objects.filter(key='autoplay').update(value=enabled)
        self.autoplay = enabled
        self._handle_autoplay()
    @disabled_when_voting
    @control
    def set_volume(self, request):
        self.volume = float(request.POST.get('value'))
        with self.mpd_command() as allowed:
            if allowed:
                try:
                    self.player.setvol(round(self.volume * 100))
                except mpd.base.CommandError:
                    # sometimes the volume can't be set
                    self.musiq.base.logger.info('could not set volume')
                    pass
    @disabled_when_voting
    @control
    def remove_all(self, request):
        if not self.musiq.base.user_manager.is_admin(request.user):
            return HttpResponseForbidden()
        with self.mpd_command() as allowed:
            if allowed:
                with transaction.atomic():
                    count = self.queue.count()
                    self.queue.all().delete()
                for _ in range(count):
                    self.queue_semaphore.acquire(blocking=False)
    @disabled_when_voting
    @control
    def prioritize(self, request):
        key = request.POST.get('key')
        if key is None:
            return HttpResponseBadRequest()
        key = int(key)
        self.queue.prioritize(key)
    @disabled_when_voting
    @control
    def remove(self, request):
        key = request.POST.get('key')
        if key is None:
            return HttpResponseBadRequest()
        key = int(key)
        try:
            removed = self.queue.remove(key)
            self.queue_semaphore.acquire(blocking=False)
            # if we removed a song and it was added by autoplay,
            # we want it to be the new basis for autoplay
            if not removed.manually_requested:
                self._handle_autoplay(removed.url)
            else:
                self._handle_autoplay()
        except models.QueuedSong.DoesNotExist:
            return HttpResponseBadRequest('song does not exist')
    @disabled_when_voting
    @control
    def reorder(self, request):
        prev = request.POST.get('prev')
        key = request.POST.get('element')
        next = request.POST.get('next')
        if key is None or len(key) == 0:
            return HttpResponseBadRequest()
        if prev is None or len(prev) == 0:
            prev = None
        else:
            prev = int(prev)
        key = int(key)
        if next is None or len(next) == 0:
            next = None
        else:
            next = int(next)
        try:
            self.queue.reorder(prev, key, next)
        except ValueError:
            return HttpResponseBadRequest('request on old state')
    @control
    def vote_up(self, request):
        key = request.POST.get('key')
        if key is None:
            return HttpResponseBadRequest()
        key = int(key)

        models.CurrentSong.objects.filter(queue_key=key).update(votes=F('votes')+1)

        self.queue.vote_up(key)
    @control
    def vote_down(self, request):
        key = request.POST.get('key')
        if key is None:
            return HttpResponseBadRequest()
        key = int(key)

        models.CurrentSong.objects.filter(queue_key=key).update(votes=F('votes')-1)
        try:
            current_song = models.CurrentSong.objects.get()
            if current_song.queue_key == key and current_song.votes <= -self.musiq.base.settings.downvotes_to_kick:
                with self.mpd_command() as allowed:
                    if allowed:
                        self.player.stop()
        except models.CurrentSong.DoesNotExist:
            pass

        removed = self.queue.vote_down(key, -self.musiq.base.settings.downvotes_to_kick)
        # if we removed a song by voting, and it was added by autoplay,
        # we want it to be the new basis for autoplay
        if removed is not None:
            self.queue_semaphore.acquire(blocking=False)
            if not removed.manually_requested:
                self._handle_autoplay(removed.url)
            else:
                self._handle_autoplay()



from django.db import models
from django.contrib import admin
import core.musiq.song_queue
import core.musiq.song_utils as song_utils

# Create your models here.
class Tag(models.Model):
    text = models.CharField(max_length=100)
    def __str__(self):
        return self.text

class Counter(models.Model):
    value = models.IntegerField()
    def __str__(self):
        return str(self.value)

class ArchivedSong(models.Model):
    url = models.CharField(max_length=200, unique=True)
    artist = models.CharField(max_length=1000)
    title = models.CharField(max_length=1000)
    counter = models.IntegerField()
    def __str__(self):
        return self.title + ' (' + self.url + '): ' + str(self.counter)
    def displayname(self):
        return song_utils.displayname(self.artist, self.title)

class ArchivedPlaylist(models.Model):
    list_id = models.CharField(max_length=200)
    title = models.CharField(max_length=1000)
    created = models.DateTimeField(auto_now_add=True)
    counter = models.IntegerField()
    def __str__(self):
        return self.title + ': ' + str(self.counter)

class PlaylistEntry(models.Model):
    playlist = models.ForeignKey('ArchivedPlaylist', on_delete=models.CASCADE, related_name='entries')
    song = models.ForeignKey('ArchivedSong', on_delete=models.SET_NULL, blank=True, null=True, related_name='playlist_entries')
    index = models.IntegerField()
    url = models.CharField(max_length=200)
    def __str__(self):
        s = self.playlist.title + '[' + str(self.index) + ']: '
        if self.song is not None:
            s += self.song.displayname()
        else:
            s += self.url
        return s
    class Meta:
       ordering = ['playlist', 'index']

class ArchivedQuery(models.Model):
    song = models.ForeignKey('ArchivedSong', on_delete=models.CASCADE, related_name='queries')
    query = models.CharField(max_length=1000)
    def __str__(self):
        return self.query

class ArchivedPlaylistQuery(models.Model):
    playlist = models.ForeignKey('ArchivedPlaylist', on_delete=models.CASCADE, related_name='queries')
    query = models.CharField(max_length=1000)
    def __str__(self):
        return self.query

class QueuedSong(models.Model):
    index = models.IntegerField()
    manually_requested = models.BooleanField()
    votes = models.IntegerField(default=0)
    url = models.CharField(max_length=200)
    artist = models.CharField(max_length=1000)
    title = models.CharField(max_length=1000)
    duration = models.CharField(max_length=20)
    objects = core.musiq.song_queue.SongQueue()
    def __str__(self):
        return str(self.index) + ': ' + self.title + ' (' + self.url + ')'
    def displayname(self):
        return song_utils.displayname(self.artist, self.title)
    class Meta:
       ordering = ['index']

class CurrentSong(models.Model):
    queue_key = models.IntegerField()
    manually_requested = models.BooleanField()
    votes = models.IntegerField()
    url = models.CharField(max_length=200)
    artist = models.CharField(max_length=1000)
    title = models.CharField(max_length=1000)
    duration = models.CharField(max_length=20)
    location = models.CharField(max_length=1000)
    created = models.DateTimeField(auto_now_add=True)
    def __str__(self):
        return self.title + ' (' + self.url + ')'
    def displayname(self):
        return song_utils.displayname(self.artist, self.title)

class RequestLog(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    song = models.ForeignKey('ArchivedSong', on_delete=models.SET_NULL, blank=True, null=True)
    playlist = models.ForeignKey('ArchivedPlaylist', on_delete=models.SET_NULL, blank=True, null=True)
    address = models.CharField(max_length=50)
    def __str__(self):
        if self.song is not None:
            return self.address + ': ' + self.song.displayname()
        elif self.playlist is not None:
            return self.address + ': ' + self.playlist.title
        else:
            return self.address + ': <None>'

class PlayLog(models.Model):
    created = models.DateTimeField(auto_now_add=True)
    song = models.ForeignKey('ArchivedSong', on_delete=models.SET_NULL, blank=True, null=True)
    manually_requested = models.BooleanField()
    votes = models.IntegerField(null=True)
    def __str__(self):
        return 'played ' + self.song.displayname() + ' with ' + str(self.votes) + ' votes'

class Setting(models.Model):
    key = models.CharField(max_length=200, unique=True)
    value = models.CharField(max_length=200)
    def __str__(self):
        return self.key + ': ' + ('None' if self.value is None else self.value)

class Pad(models.Model):
    version = models.IntegerField(default=0)
    content = models.CharField(max_length=100000)
    def __str__(self):
        return '{' + str(self.version) + '}: ' + self.content[:20] + '...'



import os

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

try:
    with open(os.path.join(BASE_DIR, 'config/secret_key.txt')) as f:
        SECRET_KEY = f.read().strip()
except FileNotFoundError:
    from django.core.management.utils import get_random_secret_key
    SECRET_KEY = get_random_secret_key()
    with open(os.path.join(BASE_DIR, 'config/secret_key.txt'), 'w') as f:
        f.write(SECRET_KEY)
    print('created secret key')

if os.environ.get('DJANGO_DEBUG'):
    DEBUG = True
else:
    DEBUG = False

ALLOWED_HOSTS = ['*']


# Application definition

INSTALLED_APPS = [
    'core.apps.CoreConfig',
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'compressor',
    'sass_processor',
    'channels',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'core.user_manager.SimpleMiddleware',
]

ROOT_URLCONF = 'main.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates/'),
            ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = 'main.wsgi.application'


# Database

if DEBUG:
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
            'OPTIONS': {
                'timeout': 20,
            }
        }
    }
else:
    DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.postgresql',
            'NAME': 'raveberry',
            'USER': 'raveberry',
            'PASSWORD': 'raveberry',
            'HOST': '127.0.0.1',
            'PORT': '5432',
        }
    }

# Password validation
AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]


# Internationalization
LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'Europe/Berlin'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Users
LOGIN_REDIRECT_URL = '/'
LOGOUT_REDIRECT_URL = '/'
# only preserve user sessions for an hour
#SESSION_COOKIE_AGE = 3600

# Static files (CSS, JavaScript, Images)
STATIC_ROOT = os.path.join(BASE_DIR, 'static')
STATICFILES_DIRS = [
        #os.path.join(BASE_DIR, "staticfiles/"),
        ]
STATIC_URL = '/static/'

# create symlink to admin static files if not present
if not os.path.islink(os.path.join(BASE_DIR, 'static/admin')):
    import django
    django_path = os.path.dirname(django.__file__)
    static_admin = os.path.join(django_path, 'contrib/admin/static/admin')
    os.symlink(static_admin, os.path.join(BASE_DIR, 'static/admin'), target_is_directory=True)
    print('linked static admin files')


# adapted for django-compressor and sass-processor
SASS_PROCESSOR_ROOT = STATIC_ROOT

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    # other finders..
    'compressor.finders.CompressorFinder',
    'sass_processor.finders.CssFinder',
)

# channels
ASGI_APPLICATION = "main.routing.application"
CHANNEL_LAYERS = {
    'default': {
        'BACKEND': 'channels_redis.core.RedisChannelLayer',
        'CONFIG': {
            "hosts": [('127.0.0.1', 6379)],
        },
    },
}

# Logging

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'infofile': {
            'level': 'INFO',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': os.path.join(BASE_DIR, 'logs/info.log'),
            'maxBytes': 1024*1024*15, # 15MB
            'backupCount': 10,
        },
        'errorfile': {
            'level': 'ERROR',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': os.path.join(BASE_DIR, 'logs/error.log'),
            'maxBytes': 1024*1024*15, # 15MB
            'backupCount': 10,
        },
        'console': {
            'class': 'logging.StreamHandler',
        },
    },
    'loggers': {
        'raveberry': {
            'handlers': ['infofile', 'errorfile'],
            'level': 'INFO',
            'propagate': True,
        },
        'django.request': {
            'handlers': ['errorfile'],
            'level': 'ERROR',
            'propagate': True,
        },
    },
}

# Security Settings
#SECURE_CONTENT_TYPE_NOSNIFF = True
#SECURE_BROWSER_XSS_FILTER = True
#SESSION_COOKIE_SECURE = True
#CSRF_COOKIE_SECURE = True
#X_FRAME_OPTIONS = 'DENY'

# app settings
try:
    with open(os.path.join(BASE_DIR, 'config/cache_dir')) as f:
        SONGS_CACHE_DIR = f.read().strip()
except FileNotFoundError:
    SONGS_CACHE_DIR = ''
if SONGS_CACHE_DIR == '':
    SONGS_CACHE_DIR = os.path.expanduser('~/Music/raveberry/')
    with open(os.path.join(BASE_DIR, 'config/cache_dir'), 'w') as f:
        f.write(SONGS_CACHE_DIR)
    print('no song caching directory specified, using ~/Music/raveberry/')



# Generated by Django 2.2.3 on 2019-07-20 23:31

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='ArchivedSong',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.CharField(max_length=200, unique=True)),
                ('artist', models.CharField(max_length=1000)),
                ('title', models.CharField(max_length=1000)),
                ('counter', models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='Counter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='CurrentSong',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('queue_key', models.IntegerField()),
                ('votes', models.IntegerField()),
                ('url', models.CharField(max_length=200)),
                ('artist', models.CharField(max_length=1000)),
                ('title', models.CharField(max_length=1000)),
                ('duration', models.CharField(max_length=20)),
                ('location', models.CharField(max_length=1000)),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name='Pad',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', models.IntegerField(default=0)),
                ('content', models.CharField(max_length=100000)),
            ],
        ),
        migrations.CreateModel(
            name='QueuedSong',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('index', models.IntegerField()),
                ('votes', models.IntegerField(default=0)),
                ('url', models.CharField(max_length=200)),
                ('artist', models.CharField(max_length=1000)),
                ('title', models.CharField(max_length=1000)),
                ('duration', models.CharField(max_length=20)),
            ],
            options={
                'ordering': ['index'],
            },
        ),
        migrations.CreateModel(
            name='Setting',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(max_length=200, unique=True)),
                ('value', models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='RequestLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('address', models.CharField(max_length=50)),
                ('song', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.ArchivedSong')),
            ],
        ),
        migrations.CreateModel(
            name='PlayLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('votes', models.IntegerField(null=True)),
                ('song', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='core.ArchivedSong')),
            ],
        ),
        migrations.CreateModel(
            name='ArchivedQuery',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('query', models.CharField(max_length=1000)),
                ('song', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='queries', to='core.ArchivedSong')),
            ],
        ),
    ]



# Generated by Django 2.2.3 on 2019-09-13 15:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0004_archivedplaylist_created'),
    ]

    operations = [
        migrations.AlterField(
            model_name='archivedplaylist',
            name='list_id',
            field=models.CharField(max_length=200),
        ),
    ]



import rpi_ws281x 

class Ring:

    def __init__(self):
        # LED ring configuration:
        LED_COUNT      = 16      # Number of LED pixels.
        LED_PIN        = 10      # GPIO pin connected to the pixels (10: SPI, 18: PWM (used for sound).
        LED_FREQ_HZ    = 800000  # LED signal frequency in hertz (usually 800khz)
        LED_DMA        = 10      # DMA channel to use for generating signal (try 10)
        LED_BRIGHTNESS = 255     # Set to 0 for darkest and 255 for brightest
        LED_INVERT     = False   # True to invert the signal (when using NPN transistor level shift)
        LED_CHANNEL    = 0       # set to '1' for GPIOs 13, 19, 41, 45 or 53

        self.LED_COUNT = LED_COUNT
        self.LED_OFFSET = 12
        self.brightness = 1
        self.monochrome = False

        self.controller = rpi_ws281x.Adafruit_NeoPixel(LED_COUNT, LED_PIN, LED_FREQ_HZ, LED_DMA, LED_INVERT, LED_BRIGHTNESS, LED_CHANNEL) 
        try:
            self.controller.begin()
            self.initialized = True
        except RuntimeError:
            # could not connect to led ring
            self.initialized = False

    def set_colors(self, colors):
        if not self.initialized:
            return
        for led in range(self.LED_COUNT):
            dimmed_color = (self.brightness * val for val in colors[led])
            scaled_color = tuple(int(val * 255) for val in dimmed_color)
            self.controller.setPixelColorRGB((led + self.LED_OFFSET) % self.LED_COUNT, *scaled_color)
        self.controller.show()

    def clear(self):
        if not self.initialized:
            return
        for led in range(self.LED_COUNT):
            self.controller.setPixelColorRGB(led, 0, 0, 0)
        self.controller.show()



# hack to allow direct calling of this script
if __name__ == '__main__':
    import sys
    sys.path.insert(0,'../../../')

from django.conf import settings

from core.lights.programs import VizProgram

import os
import sys
import time
import math
import random
import ctypes
import subprocess

class Circle(VizProgram):

    def __init__(self, lights):
        super().__init__(lights)
        self.name = 'Circular'
        self.cava = lights.cava_program

        # To have pi3d display shader compilation errors, set the logging level to INFO
        #import logging
        #logging.basicConfig(level=logging.INFO)

        # The appearance of the shader was taken from
        # https://www.shadertoy.com/view/llycWD

        # The width and height of the displayed visualizer in pixels
        self.WIDTH = 640
        self.HEIGHT = 480

        # The number of ffts that are kept in history
        self.FFT_HIST = 32
        # The part of the spectrum that is used to detect loud bass, making the circle bigger
        self.BASS_MAX = 0.05
        # The number of bins we cut after smoothing on both sides
        self.SPECTRUM_CUT = 2
        # The amount of particles
        self.NUM_PARTICLES = 400
        # Size of the particles
        self.PARTICLE_SIZE = 0.01
        # How far away to spawn particles
        self.PARTICLE_SPAWN_Z = 2.0

        self.should_prepare = False

        # https://stackoverflow.com/questions/48472285/create-core-context-with-glut-freeglut
        # In order to use newer OpenGL versions (e.g. for Instancing), use the following line:
        # glutInitContextVersion( 3, 3 );

    def set_resolution(self, width, height):
        self.WIDTH, self.HEIGHT = width, height
        # this scale value determines how many times bigger the actual display is in comparison to the actually computed buffer. Higher values equal lower computation but also lower resolution.
        # 1366x768 works alright, use it to compute a starting SCALE (in multiples of SCALE_STEP)
        self.SCALE_STEP = 0.5
        self.SCALE = round(width * height / (1366 * 768) * 1/self.SCALE_STEP) / (1/self.SCALE_STEP)
        self.SCALE = max(self.SCALE, 1)

    def start(self):
        self.cava.use()

        os.environ['DISPLAY'] = ':0'

        # disable blanking and power saving
        subprocess.call('xset s off'.split())
        subprocess.call('xset s noblank'.split())
        subprocess.call('xset -dpms'.split())

        # read the resolution from the X window and set our dimensions accordingly. Uses the DISPLAY environ variable
        # alternative without X would be tvserice -d edid.dat && edidparser edid.dat
        xwininfo = subprocess.check_output('xwininfo -root'.split()).decode().splitlines()
        width, height = 640, 480
        for line in xwininfo:
            line = line.strip()
            if line.startswith('Width:'):
                width = int(line.split()[1])
            if line.startswith('Height:'):
                height = int(line.split()[1])
        #width, height = 640, 360
        self.set_resolution(width, height)

        self.resolution_increases = {}

        self.should_prepare = True

        self.history_toggle = True

    def increase_resolution(self):
        if self.SCALE == 1:
            # do not render more than one pixel per pixel
            return

        # if we get told to increase at a specific resolution a lot of times, we got probably caught in an oscillation. Don't increase resolution then
        scale_key = round(self.SCALE, 1)
        if scale_key not in self.resolution_increases:
            self.resolution_increases[scale_key] = 0
        if self.resolution_increases[scale_key] >= 2:
            return
        self.resolution_increases[scale_key] += 1

        self.SCALE = max(1, self.SCALE - self.SCALE_STEP)

    def decrease_resolution(self):
        self.SCALE = self.SCALE + self.SCALE_STEP
    
    def _prepare(self):
        self.should_prepare = False

        import numpy as np
        # pi3d has to be imported in the same thread that it will draw in
        # Thus, import it here instead of at the top of the file
        import pi3d
        from pi3d.constants import opengles, GL_CLAMP_TO_EDGE, GL_ALWAYS, GL_NEVER, GL_DEPTH_TEST, GL_VERSION
        # used for reimplementing the draw call with instancing
        from pi3d.constants import GLsizei, GLint, GLboolean, GL_FLOAT, GL_UNSIGNED_SHORT, GLuint, GL_ARRAY_BUFFER, GL_STATIC_DRAW, GLsizeiptr, GLfloat
        from PIL import Image

        # Setup display and initialise pi3d
        self.display = pi3d.Display.create(w=self.WIDTH, h=self.HEIGHT, window_title="Raveberry")
        # error 0x500 after Display create
        error = opengles.glGetError()
        # Set a pink background color so mistakes are clearly visible
        #self.display.set_background(1, 0, 1, 1)
        self.display.set_background(0, 0, 0, 1)

        # print OpenGL Version, useful for debugging
        #import ctypes
        #def print_char_p(addr):
        #    g = (ctypes.c_char*32).from_address(addr)
        #    i = 0
        #    while True:
        #        c = g[i]
        #        if c == b'\x00':
        #            break
        #        sys.stdout.write(c.decode())
        #        i += 1
        #    sys.stdout.write('\n')
        #print_char_p(opengles.glGetString(GL_VERSION))

        '''
        Visualization is split into five parts:
        The background, the particles, the spectrum, the logo and after effects.
        * The background is a vertical gradient that cycles through HSV color space, speeding up with strong bass.
        * Particles are multiple sprites that are created at a specified x,y-coordinate and fly towards the camera.
          Due to the projection into screenspace they seem to move away from the center.
        * The spectrum is a white circle that represents the fft-transformation of the currently played audio. It is smoothed to avoid strong spikes.
        * The logo is a black circle on top of the spectrum containing the logo.
        * After effects add a vignette.
        Each of these parts is represented with pi3d Shapes. They have their own shader and are ordered on the z-axis to ensure correct overlapping.
        '''

        background_shader = pi3d.Shader(os.path.join(settings.BASE_DIR, 'core/lights/circle/background'))
        self.background = pi3d.Sprite(w=2, h=2)
        self.background.set_shader(background_shader)
        self.background.positionZ(0)

        self.particle_shader = pi3d.Shader(os.path.join(settings.BASE_DIR, 'core/lights/circle/particle'))

        # create one sprite for all particles and an array containing the position and speed for all of them
        self.particle_sprite = pi3d.Sprite(w=self.PARTICLE_SIZE, h=self.PARTICLE_SIZE)
        self.particle_sprite.set_shader(self.particle_shader)
        self.particle_sprite.positionZ(0)
        particles = self._initial_particles()

        # This part was modified from https://learnopengl.com/Advanced-OpenGL/Instancing
        self.instance_vbo = GLuint()
        opengles.glGenBuffers(GLsizei(1), ctypes.byref(self.instance_vbo))
        opengles.glBindBuffer(GL_ARRAY_BUFFER, self.instance_vbo)
        particles_raw = particles.ctypes.data_as(ctypes.POINTER(GLfloat))
        opengles.glBufferData(GL_ARRAY_BUFFER, particles.nbytes, particles_raw, GL_STATIC_DRAW)
        opengles.glBindBuffer(GL_ARRAY_BUFFER, GLuint(0))

        attr_particle = opengles.glGetAttribLocation(self.particle_shader.program, b'particle')
        opengles.glEnableVertexAttribArray(attr_particle)
        opengles.glBindBuffer(GL_ARRAY_BUFFER, self.instance_vbo)
        opengles.glVertexAttribPointer(attr_particle, GLint(4), GL_FLOAT, GLboolean(0), 0, 0)
        opengles.glBindBuffer(GL_ARRAY_BUFFER, GLuint(0))
        opengles.glVertexAttribDivisor(attr_particle, GLuint(1))

        spectrum_shader = pi3d.Shader(os.path.join(settings.BASE_DIR, 'core/lights/circle/spectrum'))

        # use the ratio to compute small sizes for the sprites
        ratio = self.WIDTH / self.HEIGHT
        self.spectrum = pi3d.Sprite(w=2/ratio, h=2)
        self.spectrum.set_shader(spectrum_shader)
        self.spectrum.positionZ(0)

        # initialize the spectogram history with zeroes
        self.fft = np.zeros((self.FFT_HIST, self.cava.bars - 2*self.SPECTRUM_CUT), dtype=np.uint8)

        logo_shader = pi3d.Shader(os.path.join(settings.BASE_DIR, 'core/lights/circle/logo'))
        self.logo = pi3d.Sprite(w=1.375/ratio, h=1.375)
        self.logo.set_shader(logo_shader)
        self.logo.positionZ(0)

        logo_image = Image.open(os.path.join(settings.BASE_DIR, 'static/graphics/raveberry_square.png'))
        self.logo_array = np.frombuffer(logo_image.tobytes(), dtype=np.uint8)
        self.logo_array = self.logo_array.reshape((logo_image.size[1], logo_image.size[0], 3))
        # add space for the spectrum
        self.logo_array = np.concatenate((self.logo_array, np.zeros((self.FFT_HIST, logo_image.size[0], 3), dtype=np.uint8)), axis=0)
        # add alpha channel
        self.logo_array = np.concatenate((self.logo_array, np.ones((self.logo_array.shape[0], self.logo_array.shape[1], 1), dtype=np.uint8)), axis=2)

        # In order to save memory, the logo and the spectrum share one texture. The upper 256x256 pixels are the raveberry logo.
        # Below are 256xFFT_HIST pixels for the spectrum. The lower part is periodically updated every frame while the logo stays static.
        self.dynamic_texture = pi3d.Texture(self.logo_array)
        # Prevent interpolation from opposite edge
        self.dynamic_texture.m_repeat = GL_CLAMP_TO_EDGE
        self.spectrum.set_textures([self.dynamic_texture])
        self.logo.set_textures([self.dynamic_texture])

        after_shader = pi3d.Shader(os.path.join(settings.BASE_DIR, 'core/lights/circle/after'))
        self.after = pi3d.Sprite(w=2, h=2)
        self.after.set_shader(after_shader)
        self.after.positionZ(0)

        # create an OffscreenTexture to allow scaling. By first rendering into a smaller Texture a lot of computation is saved. This OffscreenTexture is then drawn at the end of the draw loop.
        self.post = pi3d.util.OffScreenTexture.OffScreenTexture('scale')
        self.post_sprite = pi3d.Sprite(w=2, h=2)
        post_shader = pi3d.Shader(os.path.join(settings.BASE_DIR, 'core/lights/circle/scale'))
        self.post_sprite.set_shader(post_shader)
        self.post_sprite.set_textures([self.post])

        self.total_bass = 0
        self.last_loop = time.time()
        self.time_elapsed = 0

        opengles.glDepthFunc(GL_ALWAYS)

    def draw(self):
        import numpy as np
        from scipy.ndimage.filters import gaussian_filter
        import pi3d
        from pi3d.Camera import Camera
        from pi3d.constants import opengles, GL_LESS, GL_ALWAYS, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GLsizei, GLboolean, GLint, GL_FLOAT, GL_ARRAY_BUFFER, GL_UNSIGNED_SHORT, GL_TEXTURE_2D, GL_UNSIGNED_BYTE

        time_logging = False

        if self.should_prepare:
            self._prepare()

        if self.lights.alarm_program.factor != -1:
            alarm_factor = max(0.001, self.lights.alarm_program.factor)
        else:
            alarm_factor = 0

        then = time.time()

        self.display.loop_running()
        now = self.display.time
        time_delta = now - self.last_loop
        self.last_loop = now
        self.time_elapsed += time_delta

        if time_logging:
            print(f'{time.time() - then} main loop')
            then = time.time()

        # use a sliding window to smooth the spectrum with a gauss function
        # truncating does not save significant time (~3% for this step)

        new_frame = np.array(self.cava.current_frame, dtype='float32')
        new_frame = gaussian_filter(self.cava.current_frame, sigma=1.5, mode='nearest')
        new_frame = new_frame[self.SPECTRUM_CUT:-self.SPECTRUM_CUT]
        new_frame = -0.5*new_frame**3+1.5*new_frame
        new_frame *= 255
        current_frame = new_frame

        if time_logging:
            print(f'{time.time() - then} spectrum smoothing')
            then = time.time()

        # Value used for circle shake and background color cycle
        # select the first few values and compute their average
        bass_elements = math.ceil(self.BASS_MAX * self.cava.bars)
        bass_value = sum(current_frame[0:bass_elements]) / bass_elements / 255
        bass_value = max(bass_value, alarm_factor)
        self.total_bass = self.total_bass + bass_value
        # the fraction of time that there was bass
        bass_fraction = self.total_bass / self.time_elapsed / self.lights.UPS

        # start rendering to the smaller OffscreenTexture
        # we decrease the size of the texture so it only allocates that much memory
        # otherwise it would use as much as the displays size, negating its positive effect
        self.post.ix = int(self.post.ix / self.SCALE)
        self.post.iy = int(self.post.iy / self.SCALE)
        opengles.glViewport(GLint(0), GLint(0), GLsizei(int(self.WIDTH/self.SCALE)), GLsizei(int(self.HEIGHT/self.SCALE)))
        self.post._start()
        self.post.ix = self.WIDTH
        self.post.iy = self.HEIGHT

        self.background.unif[48] = self.WIDTH / self.SCALE
        self.background.unif[49] = self.HEIGHT / self.SCALE
        self.background.unif[54] = self.time_elapsed
        self.background.unif[56] = alarm_factor
        self.background.unif[58] = self.total_bass
        self.background.draw()

        if time_logging:
            print(f'{time.time() - then} background draw')
            then = time.time()

        # enable additive blending so the draw order of overlapping particles does not matter
        opengles.glBlendFunc(1, 1)

        self.particle_sprite.unif[53] = self.PARTICLE_SPAWN_Z
        self.particle_sprite.unif[54] = self.time_elapsed
        self.particle_sprite.unif[59] = bass_fraction

        # copied code from pi3d.Shape.draw()
        # we don't need modelmatrices, normals ord textures and always blend
        self.particle_sprite.load_opengl()
        camera = Camera.instance()
        if not camera.mtrx_made:
            camera.make_mtrx()
        self.particle_sprite.MRaw = self.particle_sprite.tr1
        self.particle_sprite.M[0,:,:] = self.particle_sprite.MRaw[:,:]
        self.particle_sprite.M[1,:,:] = np.dot(self.particle_sprite.MRaw, camera.mtrx)[:,:]

        # Buffer.draw()
        buf = self.particle_sprite.buf[0]
        buf.load_opengl()
        shader = buf.shader
        shader.use()
        opengles.glUniformMatrix4fv(shader.unif_modelviewmatrix, GLsizei(2), GLboolean(0), self.particle_sprite.M.ctypes.data)
        opengles.glUniform3fv(shader.unif_unif, GLsizei(20), self.particle_sprite.unif)
        buf._select()
        opengles.glVertexAttribPointer(shader.attr_vertex, GLint(3), GL_FLOAT, GLboolean(0), buf.N_BYTES, 0)
        opengles.glEnableVertexAttribArray(shader.attr_vertex)
        opengles.glVertexAttribPointer(shader.attr_texcoord, GLint(2), GL_FLOAT, GLboolean(0), buf.N_BYTES, 24)
        opengles.glEnableVertexAttribArray(shader.attr_texcoord)
        buf.disp.last_shader = shader
        opengles.glUniform3fv(shader.unif_unib, GLsizei(5), buf.unib)

        opengles.glBindBuffer(GL_ARRAY_BUFFER, self.instance_vbo);
        opengles.glDrawElementsInstanced(buf.draw_method, GLsizei(buf.ntris * 3), GL_UNSIGNED_SHORT, 0, self.NUM_PARTICLES)

        # restore normal blending
        opengles.glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

        if time_logging:
            print(f'{time.time() - then} particle draw')
            then = time.time()

        # roll the history one further, insert the current one. 
        # we use a texture with four channels eventhough we only need one, refer to this post:
        # https://community.khronos.org/t/updating-textures-per-frame/75020/3
        # basically the gpu converts it anyway, so other formats would be slower
        history = np.zeros((self.FFT_HIST, self.cava.bars - 2*self.SPECTRUM_CUT, 4), dtype='uint8')
        self.fft = np.roll(self.fft, 1, 0)
        self.fft[0] = current_frame
        history[:,:,0] = self.fft

        if time_logging:
            print(f'{time.time() - then} spectrum roll')
            then = time.time()

        # change the spectrum part of the texture (the lower 256xFFT_HIST pixels)
        opengles.glBindTexture(GL_TEXTURE_2D, self.dynamic_texture._tex)
        iformat = self.dynamic_texture._get_format_from_array(history, self.dynamic_texture.i_format)
        opengles.glTexSubImage2D(GL_TEXTURE_2D, 0, 0, self.dynamic_texture.ix, history.shape[1], history.shape[0], iformat, GL_UNSIGNED_BYTE, history.ctypes.data_as(ctypes.POINTER(ctypes.c_ubyte)))


        if time_logging:
            print(f'{time.time() - then} glTexImage2D')
            then = time.time()

        self.spectrum.unif[48] = self.WIDTH / self.SCALE
        self.spectrum.unif[49] = self.HEIGHT / self.SCALE
        self.spectrum.unif[51] = self.FFT_HIST
        self.spectrum.unif[52] = self.NUM_PARTICLES
        self.spectrum.unif[53] = self.PARTICLE_SPAWN_Z
        self.spectrum.unif[54] = self.time_elapsed
        self.spectrum.unif[55] = time_delta
        self.spectrum.unif[57] = bass_value
        self.spectrum.unif[58] = self.total_bass
        self.spectrum.draw()

        if time_logging:
            print(f'{time.time() - then} spectrum draw')
            then = time.time()

        self.logo.unif[48] = self.WIDTH / self.SCALE
        self.logo.unif[49] = self.HEIGHT / self.SCALE
        self.logo.unif[51] = self.FFT_HIST
        self.logo.unif[54] = self.time_elapsed
        self.logo.unif[57] = bass_value
        self.logo.unif[58] = self.total_bass
        self.logo.draw()

        if time_logging:
            print(f'{time.time() - then} logo draw')
            then = time.time()

        self.after.unif[48] = self.WIDTH / self.SCALE
        self.after.unif[49] = self.HEIGHT / self.SCALE
        self.after.unif[54] = self.time_elapsed
        self.after.unif[57] = bass_value
        self.after.draw()

        if time_logging:
            print(f'{time.time() - then} after draw')
            then = time.time()

        self.post._end()

        opengles.glViewport(GLint(0), GLint(0), GLsizei(self.WIDTH), GLsizei(self.HEIGHT))
        self.post_sprite.unif[50] = self.SCALE
        self.post_sprite.draw()

        if time_logging:
            print(f'{time.time() - then} post draw')
            then = time.time()
            print(f'scale: {self.SCALE}')
            print('=====')

    def _initial_particles(self):
        ''' constructs an array of particles containing x, y and speed value for each particle '''
        import numpy as np
        particles = np.zeros((self.NUM_PARTICLES, 4), dtype='float32')
        for index in range(particles.shape[0]):
            # Generate random particle
            phi = random.random() * 2 * math.pi
            radius_diff = random.random()
            x = math.cos(phi) * (0.2 + radius_diff * 0.05)
            y = math.sin(phi) * (0.15 + radius_diff * 0.05)
            z = self.PARTICLE_SPAWN_Z
            speed = 0.15 * (random.random() * 0.75 + 0.3)
            offset = random.random() * self.PARTICLE_SPAWN_Z

            particles[index] = [x, y, speed, offset]
        return particles

    def stop(self):
        # clean up the display. after stopping the main loop has to be called once more
        self.display.stop()
        self.display.loop_running()
        self.display.destroy()
        self.cava.release()

if __name__ == '__main__':
    cava = type('obj', (object,), {
        'bars' : 199,
        'current_frame': [0 for i in range(199)],
        'use': lambda:...,
        'release': lambda:...
    })
    lights = type('obj', (object,), {
        'UPS': 25,
        'cava_program': cava,
        'alarm_program': type('obj', (object,), {'factor': -1}),
    })
    settings = type('obj', (object,), {'BASE_DIR' : '../../..'})
    circle = Circle(lights)
    circle.start()
    seconds_per_frame = 1 / lights.UPS

    while True:
        computation_start = time.time()
        circle.draw()
        cava.current_frame = [0.5*(1+math.sin(-4*circle.time_elapsed+0.5*i)) for i in range(len(cava.current_frame))]
        cava.current_frame = [0.8 * 0.5*(1+math.sin(4*circle.time_elapsed)) * 0.5*(1+math.sin(-4*circle.time_elapsed+0.2*i)) for i in range(len(cava.current_frame))]
        computation_time = time.time() - computation_start
        #print(f'time needed / avaliable: {computation_time / seconds_per_frame:0.2f}')
        try:
            time.sleep(seconds_per_frame - computation_time)
        except ValueError:
            pass



from channels.routing import ProtocolTypeRouter, URLRouter
from channels.auth import AuthMiddlewareStack
import core.routing

application = ProtocolTypeRouter({
    'websocket': AuthMiddlewareStack(
        URLRouter(
            core.routing.websocket_urlpatterns
        )
    ),
})



from channels.generic.websocket import WebsocketConsumer
from channels.layers import get_channel_layer
from asgiref.sync import async_to_sync

import json
import time

from threading import Thread

def update_state(state):
    data = {
        'type': 'state_update',
        'state': state,
    }
    channel_layer = get_channel_layer()
    async_to_sync(channel_layer.group_send)(
        'state',
        data
    )

class StateConsumer(WebsocketConsumer):
    def connect(self):
        async_to_sync(self.channel_layer.group_add)('state', self.channel_name)
        self.accept()

    def disconnect(self, close_code):
        async_to_sync(self.channel_layer.group_discard)('state', self.channel_name)

    def receive(self, text_data):
        pass

    # Receive message from room group
    def state_update(self, event):
        # Send message to WebSocket
        self.send(text_data=json.dumps(event['state']))



from django.urls import include
from django.urls import path
from django.contrib import admin
from django.views.generic import RedirectView

import os

if os.environ.get('DJANGO_MOCK'):
    import core.mock
    urlpatterns = [
        path('', core.mock.index),
    ]
else:
    from core.base import Base
    base = Base()

    urlpatterns = [
        path('', RedirectView.as_view(pattern_name='musiq', permanent=False), name='base'),

        path('musiq/', base.musiq.index, name='musiq'),
        path('lights/', base.lights.index, name='lights'),
        path('pad/', base.pad.index, name='pad'),
        path('settings/', base.settings.index, name='settings'),
        path('accounts/', include('django.contrib.auth.urls')),
        path('login/', RedirectView.as_view(pattern_name='login', permanent=False)),
        path('logout/', RedirectView.as_view(pattern_name='logout', permanent=False)),

        path('ajax/', include([
            path('state/', base.get_state, name='base_state'),
            path('submit_hashtag/', base.submit_hashtag, name='submit_hashtag'),
            path('musiq/', include([
                path('state/', base.musiq.get_state, name='musiq_state'),
                path('random_suggestion/', base.musiq.song_provider.random_suggestion, name='random_suggestion'),
                path('request_archived_music/', base.musiq.request_archived_music, name='request_archived_music'),
                path('request_new_music/', base.musiq.request_new_music, name='request_new_music'),
                path('suggestions/', base.musiq.song_provider.get_suggestions, name='suggestions'),

                path('restart/', base.musiq.player.restart, name='restart_song'),
                path('seek_backward/', base.musiq.player.seek_backward, name='seek_backward'),
                path('play/', base.musiq.player.play, name='play_song'),
                path('pause/', base.musiq.player.pause, name='pause_song'),
                path('seek_forward/', base.musiq.player.seek_forward, name='seek_forward'),
                path('skip/', base.musiq.player.skip, name='skip_song'),
                path('set_shuffle/', base.musiq.player.set_shuffle, name='set_shuffle'),
                path('set_repeat/', base.musiq.player.set_repeat, name='set_repeat'),
                path('set_autoplay/', base.musiq.player.set_autoplay, name='set_autoplay'),
                path('request_radio/', base.musiq.request_radio, name='request_radio'),
                path('set_volume/', base.musiq.player.set_volume, name='set_volume'),
                path('remove_all/', base.musiq.player.remove_all, name='remove_all'),

                path('prioritize/', base.musiq.player.prioritize, name='prioritize_song'),
                path('remove/', base.musiq.player.remove, name='remove_song'),
                path('reorder/', base.musiq.player.reorder, name='reorder_song'),
                path('vote_up/', base.musiq.player.vote_up, name='vote_up_song'),
                path('vote_down/', base.musiq.player.vote_down, name='vote_down_song'),
            ])),
            path('lights/', include([
                path('state/', base.lights.get_state, name='lights_state'),
                path('shortcut/', base.lights.set_lights_shortcut, name='set_lights_shortcut'),
                path('set_ring_program/', base.lights.set_ring_program, name='set_ring_program'),
                path('set_ring_brightness/', base.lights.set_ring_brightness, name='set_ring_brightness'),
                path('set_ring_monochrome/', base.lights.set_ring_monochrome, name='set_ring_monochrome'),
                path('set_strip_program/', base.lights.set_strip_program, name='set_strip_program'),
                path('set_strip_brightness/', base.lights.set_strip_brightness, name='set_strip_brightness'),
                path('adjust_screen/', base.lights.adjust_screen, name='adjust_screen'),
                path('set_screen_program/', base.lights.set_screen_program, name='set_screen_program'),
                path('set_program_speed/', base.lights.set_program_speed, name='set_program_speed'),
                path('set_fixed_color/', base.lights.set_fixed_color, name='set_fixed_color'),
            ])),
            path('pad/', include([
                path('state/', base.pad.get_state, name='pad_state'),
                path('submit/', base.pad.submit, name='submit_pad'),
            ])),
            path('settings/', include([
                path('state/', base.settings.get_state, name='settings_state'),
                path('set_voting_system/', base.settings.set_voting_system, name='set_voting_system'),
                path('set_logging_enabled/', base.settings.set_logging_enabled, name='set_logging_enabled'),
                path('set_people_to_party/', base.settings.set_people_to_party, name='set_people_to_party'),
                path('set_alarm_probability/', base.settings.set_alarm_probability, name='set_alarm_probability'),
                path('set_downvotes_to_kick/', base.settings.set_downvotes_to_kick, name='set_downvotes_to_kick'),
                path('set_max_download_size/', base.settings.set_max_download_size, name='set_max_download_size'),
                path('set_max_playlist_items/', base.settings.set_max_playlist_items, name='set_max_playlist_items'),
                path('check_internet/', base.settings.check_internet, name='check_internet'),
                path('update_user_count/', base.settings.update_user_count, name='update_user_count'),

                path('set_bluetooth_scanning/', base.settings.set_bluetooth_scanning, name='set_bluetooth_scanning'),
                path('connect_to_bluetooth_device/', base.settings.connect_to_bluetooth_device, name='connect_to_bluetooth_device'),

                path('available_ssids/', base.settings.available_ssids, name='available_ssids'),
                path('connect_to_wifi/', base.settings.connect_to_wifi, name='connect_to_wifi'),
                path('enable_homewifi/', base.settings.enable_homewifi, name='enable_homewifi'),
                path('disable_homewifi/', base.settings.disable_homewifi, name='disable_homewifi'),
                path('stored_ssids/', base.settings.stored_ssids, name='stored_ssids'),
                path('set_homewifi_ssid/', base.settings.set_homewifi_ssid, name='set_homewifi_ssid'),

                path('analyse/', base.settings.analyse, name='analyse'),

                path('disable_events/', base.settings.disable_events, name='disable_events'),
                path('enable_events/', base.settings.enable_events, name='enable_events'),
                path('disable_hotspot/', base.settings.disable_hotspot, name='disable_hotspot'),
                path('enable_hotspot/', base.settings.enable_hotspot, name='enable_hotspot'),
                path('unprotect_wifi/', base.settings.unprotect_wifi, name='unprotect_wifi'),
                path('protect_wifi/', base.settings.protect_wifi, name='protect_wifi'),
                path('disable_tunneling/', base.settings.disable_tunneling, name='disable_tunneling'),
                path('enable_tunneling/', base.settings.enable_tunneling, name='enable_tunneling'),
                path('disable_remote/', base.settings.disable_remote, name='disable_remote'),
                path('enable_remote/', base.settings.enable_remote, name='enable_remote'),
                path('reboot_server/', base.settings.reboot_server, name='reboot_server'),
                path('reboot_system/', base.settings.reboot_system, name='reboot_system'),
            ])),
        ])),
    ]
